# -*- coding: utf-8 -*-
import logging
import operator
import shutil
import tempfile
from pathlib import Path
from typing import Optional
from uuid import UUID

from apistar.exceptions import ErrorResponse
from arkindex_export import open_database
from arkindex_worker.cache import (
    CachedClassification,
    CachedElement,
    CachedEntity,
    CachedImage,
    CachedTranscription,
    CachedTranscriptionEntity,
    create_tables,
    create_version_table,
)
from arkindex_worker.cache import db as cache_database
from arkindex_worker.cache import init_cache_db
from arkindex_worker.worker.base import BaseWorker
from worker_generic_training_dataset.db import (
    get_children,
    list_classifications,
    list_transcriptions,
    retrieve_element,
    retrieve_entities,
)
from worker_generic_training_dataset.utils import (
    create_tar_zstd_archive,
    download_image,
)

logger = logging.getLogger(__name__)

BULK_BATCH_SIZE = 50


class DatasetExtractor(BaseWorker):
    def configure(self):
        self.args = self.parser.parse_args()
        if self.is_read_only:
            super().configure_for_developers()
        else:
            super().configure()

        if self.user_configuration:
            logger.info("Overriding with user_configuration")
            self.config.update(self.user_configuration)

        # database arg is mandatory in dev mode
        assert (
            not self.is_read_only or self.args.database is not None
        ), "`--database` arg is mandatory in developer mode."

        # Read process information
        self.read_training_related_information()

        # Download corpus
        self.download_latest_export()

        # Initialize db that will be written
        self.initialize_database()

        # Cached Images downloaded and created in DB
        self.cached_images = dict()

    def read_training_related_information(self):
        """
        Read from process information
            - train_folder_id
            - validation_folder_id
            - test_folder_id (optional)
            - model_id
        """
        logger.info("Retrieving information from process_information")

        train_folder_id = self.config.get("train_folder_id")
        assert train_folder_id, "A training folder id is necessary to use this worker"
        self.training_folder_id = UUID(train_folder_id)

        val_folder_id = self.config.get("validation_folder_id")
        assert val_folder_id, "A validation folder id is necessary to use this worker"
        self.validation_folder_id = UUID(val_folder_id)

        test_folder_id = self.config.get("test_folder_id")
        self.testing_folder_id = UUID(test_folder_id) if test_folder_id else None

    def initialize_database(self):
        # Create db at
        # - self.workdir / "db.sqlite" in Arkindex mode
        # - self.args.database in dev mode
        database_path = (
            Path(self.args.database)
            if self.is_read_only
            else self.work_dir / "db.sqlite"
        )
        if database_path.exists():
            database_path.unlink()

        init_cache_db(database_path)

        create_version_table()

        create_tables()

    def download_latest_export(self):
        # Find export of corpus
        try:
            exports = self.api_client.request(
                "ListExports",
                id=self.corpus_id,
            )["results"]
        except ErrorResponse as e:
            logger.error(
                f"Could not list exports of corpus ({self.corpus_id}): {str(e.content)}"
            )
            raise e

        # Find latest that is in "done" state
        exports = sorted(
            list(filter(lambda exp: exp["state"] == "done", exports)),
            key=operator.itemgetter("updated"),
            reverse=True,
        )
        assert len(exports) > 0, "No available exports found."

        # Download latest it in a tmpfile
        try:
            export_id = exports[0]["id"]
            logger.info(f"Downloading export ({export_id})...")
            self.export = self.api_client.request(
                "DownloadExport",
                id=export_id,
            )
            logger.info(f"Downloaded export ({export_id}) @ `{self.export.name}`")
            open_database(self.export.name)
        except ErrorResponse as e:
            logger.error(
                f"Could not download export ({export_id}) of corpus ({self.corpus_id}): {str(e)}"
            )
            raise e

    def insert_element(
        self, element, image_folder: Path, parent_id: Optional[str] = None
    ):
        logger.info(f"Processing element ({element.id})")
        if element.image and element.image.id not in self.cached_images:
            # Download image
            logger.info("Downloading image")
            download_image(element, folder=image_folder)
            # Insert image
            logger.info("Inserting image")
            # Store images in case some other elements use it as well
            self.cached_images[element.image_id] = CachedImage.create(
                id=element.image.id,
                width=element.image.width,
                height=element.image.height,
                url=element.image.url,
            )

        # Insert element
        logger.info("Inserting element")
        cached_element = CachedElement.create(
            id=element.id,
            parent_id=parent_id,
            type=element.type,
            image=self.cached_images[element.image.id] if element.image else None,
            polygon=element.polygon,
            rotation_angle=element.rotation_angle,
            mirrored=element.mirrored,
            worker_version_id=element.worker_version
            if element.worker_version
            else None,
            confidence=element.confidence,
        )

        # Insert classifications
        logger.info("Listing classifications")
        classifications = [
            CachedClassification(
                id=classification.id,
                element=cached_element,
                class_name=classification.class_name,
                confidence=classification.confidence,
                state=classification.state,
            )
            for classification in list_classifications(element.id)
        ]
        if classifications:
            logger.info(f"Inserting {len(classifications)} classification(s)")
            with cache_database.atomic():
                CachedClassification.bulk_create(
                    model_list=classifications,
                    batch_size=BULK_BATCH_SIZE,
                )

        # Insert transcriptions
        logger.info("Listing transcriptions")
        transcriptions = list_transcriptions(cached_element)
        if transcriptions:
            logger.info(f"Inserting {len(transcriptions)} transcription(s)")
            with cache_database.atomic():
                CachedTranscription.bulk_create(
                    model_list=transcriptions,
                    batch_size=BULK_BATCH_SIZE,
                )

        logger.info("Listing entities")
        entities, transcription_entities = [], []
        for transcription in transcriptions:
            ents, transc_ents = retrieve_entities(transcription)
            entities.extend(ents)
            transcription_entities.extend(transc_ents)

        if entities:
            logger.info(f"Inserting {len(entities)} entities")
            with cache_database.atomic():
                CachedEntity.bulk_create(
                    model_list=entities,
                    batch_size=BULK_BATCH_SIZE,
                )
            # Insert transcription entities
            logger.info(
                f"Inserting {len(transcription_entities)} transcription entities"
            )
            with cache_database.atomic():
                CachedTranscriptionEntity.bulk_create(
                    model_list=transcription_entities,
                    batch_size=BULK_BATCH_SIZE,
                )

    def process_split(self, split_name, split_id, image_folder):
        logger.info(
            f"Filling the Base-Worker cache with information from children under element ({split_id})"
        )
        # Fill cache
        # Retrieve parent and create parent
        parent = retrieve_element(split_id)
        self.insert_element(parent, image_folder)

        # First list all pages
        pages = get_children(parent_id=split_id, element_type="page")
        nb_pages = pages.count()
        for idx, page in enumerate(pages, start=1):
            logger.info(f"Processing `{split_name}` page ({idx}/{nb_pages})")

            # Insert page
            self.insert_element(page, image_folder, parent_id=split_id)

            # List children
            children = get_children(parent_id=page.id)
            nb_children = children.count()
            for child_idx, child in enumerate(children, start=1):
                logger.info(f"Processing child ({child_idx}/{nb_children})")
                # Insert child
                self.insert_element(child, image_folder, parent_id=page.id)

    def run(self):
        self.configure()

        # Where to save the downloaded images
        image_folder = Path(tempfile.mkdtemp())

        # Iterate over given split
        for split_name, split_id in [
            ("Train", self.training_folder_id),
            ("Validation", self.validation_folder_id),
            ("Test", self.testing_folder_id),
        ]:
            if not split_id:
                continue
            self.process_split(split_name, split_id, image_folder)

        # TAR + ZSTD Image folder and store as task artifact
        zstd_archive_path = self.work_dir / "arkindex_data.zstd"
        logger.info(f"Compressing the images to {zstd_archive_path}")
        create_tar_zstd_archive(folder_path=image_folder, destination=zstd_archive_path)

        # Cleanup image folder
        shutil.rmtree(image_folder)


def main():
    DatasetExtractor(
        description="Fill base-worker cache with information about dataset and extract images"
    ).run()


if __name__ == "__main__":
    main()
